/*******************************************************************************
*
* E M B E D D E D   W I Z A R D   P R O J E C T
*
*                                                Copyright (c) TARA Systems GmbH
*                                    written by Paul Banach and Manfred Schweyer
*
********************************************************************************
*
* This software is delivered "as is" and shows the usage of other software
* components. It is provided as an example software which is intended to be
* modified and extended according to particular requirements.
*
* TARA Systems hereby disclaims all warranties and conditions with regard to the
* software, including all implied warranties and conditions of merchantability
* and non-infringement of any third party IPR or other rights which may result
* from the use or the inability to use the software.
*
********************************************************************************
*
* DESCRIPTION:
*   This file is part of the interface (glue layer) between an Embedded Wizard
*   generated UI application and the board support package (BSP) of a dedicated
*   target.
*   This template is responsible to initialize the display hardware of the board
*   and to provide the necessary access to update the display content.
*   The color format of the framebuffer has to correspond to the color format
*   of the Graphics Engine.
*
*   Important: This file is intended to be used as a template. Please adapt the
*   implementation according your particular hardware.
*
*******************************************************************************/

#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>

#include "ewconfig.h"
#include "ewrte.h"
#include "ewgfx.h"
#include "ewextgfx.h"
#include "ewgfxdefs.h"

#include "ew_bsp_display.h"
#include "ew_bsp_clock.h"
#include "ew_bsp_os.h"

#include "r_glcdc.h"
#include "r_display_api.h"
#include "ew_thread.h"

display_cfg_t DisplayConfig;

#if EW_USE_OPERATING_SYSTEM == 1

  static XSemaphoreHandle     LcdUpdateSemaphore = 0;

#endif

uint8_t FrameBuffer_1[ EW_FRAME_BUFFER_SIZE ] BSP_ALIGN_VARIABLE(64) BSP_PLACE_IN_SECTION(".sdram");
#define EW_FRAME_BUFFER_ADDR (unsigned long)FrameBuffer_1

#if EW_USE_DOUBLE_BUFFER == 1

  uint8_t FrameBuffer_2[ EW_FRAME_BUFFER_SIZE ] BSP_ALIGN_VARIABLE(64) BSP_PLACE_IN_SECTION(".sdram");
  #define EW_DOUBLE_BUFFER_ADDR (unsigned long)FrameBuffer_2

#endif

#if EW_USE_DOUBLE_BUFFER == 1

  static volatile uint32_t      PendingFramebuffer = 0;

#else

  static volatile unsigned long VSyncTime          = 0;
  static volatile unsigned long VSyncDuration      = 0;

#endif

void glcdc_callback (display_callback_args_t * pdata)
{
  display_callback_args_t * p_decode;
  p_decode = (display_callback_args_t *) pdata;

  if (DISPLAY_EVENT_GR1_UNDERFLOW == p_decode->event)
  {
  }
  else if (DISPLAY_EVENT_GR2_UNDERFLOW == p_decode->event)
  {
  }
  else /* DISPLAY_EVENT_LINE_DETECTION */
  {
    #if EW_USE_DOUBLE_BUFFER == 1

      if ( PendingFramebuffer )
      {
        #if EW_USE_OPERATING_SYSTEM == 1

          EwBspOsSemaphoreRelease( LcdUpdateSemaphore );

        #endif

        PendingFramebuffer = 0;
      }

    #else

      unsigned long curTicks = DWT->CYCCNT;
      VSyncDuration = curTicks - VSyncTime;
      VSyncTime = curTicks;

    #endif
  }
}


/* Helper function to sync on a certain line number updated by the display.
   This function is necessary when the system is used in single framebuffer mode. */
#if EW_USE_DOUBLE_BUFFER == 0
static void SyncOnLine( int aLine )
{
  unsigned long ticksTillRequestedLine;
  unsigned long localVSyncTime;

  /* line interrupt is configured to begin of front porch area - we have to add vertical
     front porch and vertical back porch to start on the desired visible line */
  aLine += ( DisplayConfig.output.vtiming.total_cyc - DisplayConfig.output.vtiming.display_cyc );

  /* calculate time tick that corresponds to requested line */
  ticksTillRequestedLine = aLine * ( VSyncDuration / DisplayConfig.output.vtiming.total_cyc );

  CPU_LOAD_SET_IDLE();

  /* if requested line is missed, wait for next display frame (next V-sync) */
  while (( DWT->CYCCNT - VSyncTime ) > ticksTillRequestedLine )
    ;

  /* wait for requested line in current display frame */
  localVSyncTime = VSyncTime;
  while (( DWT->CYCCNT - localVSyncTime ) < ticksTillRequestedLine )
    ;

  CPU_LOAD_SET_ACTIVE();
}
#endif


/*******************************************************************************
* FUNCTION:
*   EwBspDisplayInit
*
* DESCRIPTION:
*   The function EwBspDisplayInit initializes the display hardware and returns
*   the display parameter.
*
* ARGUMENTS:
*   aGuiWidth,
*   aGuiHeight   - Size of the GUI in pixel.
*   aDisplayInfo - Display info data structure.
*
* RETURN VALUE:
*   Returns 1 if successful, 0 otherwise.
*
*******************************************************************************/
int EwBspDisplayInit( int aGuiWidth, int aGuiHeight, XDisplayInfo* aDisplayInfo )
{
  EW_UNUSED_ARG( aGuiWidth );
  EW_UNUSED_ARG( aGuiHeight );

  /* check and clean display info structure */
  if ( !aDisplayInfo )
    return 0;
  memset( aDisplayInfo, 0, sizeof( XDisplayInfo ));

  #if EW_USE_OPERATING_SYSTEM == 1

    /* create semaphore for handling double-buffering on V-sync */
    // LcdUpdateSemaphore = EwBspOsSemaphoreCreate( 1, 1 );
    /* use static semaphore from FSP configuration */
    LcdUpdateSemaphore = g_display_semaphore;

  #endif

  /* take display settings from FSP configuration - except the framebuffer(s) */
  DisplayConfig = g_display0_cfg;
  display_in_format_t format;

  #if ( EW_FRAME_BUFFER_COLOR_FORMAT == EW_FRAME_BUFFER_COLOR_FORMAT_RGB565 )
    format = DISPLAY_IN_FORMAT_16BITS_RGB565;
  #elif ( EW_FRAME_BUFFER_COLOR_FORMAT == EW_FRAME_BUFFER_COLOR_FORMAT_Index8 )
    format = DISPLAY_IN_FORMAT_CLUT8;
  #elif ( EW_FRAME_BUFFER_COLOR_FORMAT == EW_FRAME_BUFFER_COLOR_FORMAT_LumA44 )
    format = DISPLAY_IN_FORMAT_CLUT8;
  #endif

  DisplayConfig.input[ DISPLAY_FRAME_LAYER_1 ].format  = format;
  DisplayConfig.input[ DISPLAY_FRAME_LAYER_1 ].hsize   = EW_DISPLAY_WIDTH;
  DisplayConfig.input[ DISPLAY_FRAME_LAYER_1 ].vsize   = EW_DISPLAY_HEIGHT;
  DisplayConfig.input[ DISPLAY_FRAME_LAYER_1 ].p_base  = (uint32_t*)EW_FRAME_BUFFER_ADDR;
  DisplayConfig.input[ DISPLAY_FRAME_LAYER_1 ].hstride = EW_FRAME_BUFFER_WIDTH;

  R_GLCDC_Open( &g_display0_ctrl, &DisplayConfig );
  R_GLCDC_Start( &g_display0_ctrl );

  /* return the current display configuration */
  aDisplayInfo->FrameBuffer    = (void*)EW_FRAME_BUFFER_ADDR;
  aDisplayInfo->DoubleBuffer   = (void*)EW_DOUBLE_BUFFER_ADDR;
  aDisplayInfo->BufferWidth    = EW_FRAME_BUFFER_WIDTH;
  aDisplayInfo->BufferHeight   = EW_FRAME_BUFFER_HEIGHT;
  aDisplayInfo->DisplayWidth   = EW_DISPLAY_WIDTH;
  aDisplayInfo->DisplayHeight  = EW_DISPLAY_HEIGHT;

  #if EW_USE_DOUBLE_BUFFER == 1
    aDisplayInfo->UpdateMode   = EW_BSP_DISPLAY_UPDATE_NORMAL;
  #else
    aDisplayInfo->UpdateMode   = EW_BSP_DISPLAY_UPDATE_PARTIAL;
  #endif

  /* enable the backlight */
  R_IOPORT_PinWrite( &g_ioport_ctrl, DISP_BLEN, BSP_IO_LEVEL_HIGH );

  return 1;
}


/*******************************************************************************
* FUNCTION:
*   EwBspDisplayDone
*
* DESCRIPTION:
*   The function EwBspDisplayDone deinitializes the display hardware.
*
* ARGUMENTS:
*   aDisplayInfo - Display info data structure.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspDisplayDone( XDisplayInfo* aDisplayInfo )
{
  EW_UNUSED_ARG( aDisplayInfo );
}


/*******************************************************************************
* FUNCTION:
*   EwBspDisplayGetUpdateArea
*
* DESCRIPTION:
*   The function EwBspDisplayGetUpdateArea returns the next update area
*   depending on the selected display mode:
*   In case of a synchroneous single-buffer, the function has to return the
*   the rectangular areas that correspond to the horizontal stripes (fields)
*   of the framebuffer.
*   In case of a scratch-pad buffer, the function has to return the subareas
*   that fit into the provided update rectangle.
*   During each display update, this function is called until it returns 0.
*
* ARGUMENTS:
*   aUpdateRect - Rectangular area which should be updated (redrawn).
*
* RETURN VALUE:
*   Returns 1 if a further update area can be provided, 0 otherwise.
*
*******************************************************************************/
int EwBspDisplayGetUpdateArea( XRect* aUpdateRect )
{
  #if EW_USE_DOUBLE_BUFFER == 0

    static int field = 0;

    /* check if last field is already reached */
    if ( field >= EW_NUMBER_OF_FIELDS )
    {
      field = 0;
      return 0;
    }

    /* determine rectangular area of current field */
    #if EW_SURFACE_ROTATION == 0
      *aUpdateRect = EwNewRect( 0, field * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS,
        EW_FRAME_BUFFER_WIDTH, ( field + 1 ) * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS );
    #endif

    #if EW_SURFACE_ROTATION == 90
      *aUpdateRect = EwNewRect( field * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS, 0,
        ( field + 1 ) * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS, EW_FRAME_BUFFER_WIDTH );
    #endif

    #if EW_SURFACE_ROTATION == 180
      *aUpdateRect = EwNewRect( 0, EW_FRAME_BUFFER_HEIGHT - ( field + 1 ) * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS,
        EW_FRAME_BUFFER_WIDTH, EW_FRAME_BUFFER_HEIGHT - field * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS );
    #endif

    #if EW_SURFACE_ROTATION == 270
      *aUpdateRect = EwNewRect( EW_FRAME_BUFFER_HEIGHT - ( field + 1 ) * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS,
        0, EW_FRAME_BUFFER_HEIGHT - field * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS, EW_FRAME_BUFFER_WIDTH );
    #endif

    /* next field */
    field++;

    /* sync on start line of next field to ensure save drawing operation */
    SyncOnLine( field * EW_FRAME_BUFFER_HEIGHT / EW_NUMBER_OF_FIELDS );

    return 1;

  #endif

  return 0;
}


/*******************************************************************************
* FUNCTION:
*   EwBspDisplayWaitForCompletion
*
* DESCRIPTION:
*   The function EwBspDisplayWaitForCompletion is called from the Graphics Engine
*   to ensure that all pending activities of the display system are completed, so
*   that the rendering of the next frame can start.
*   In case of a double-buffering system, the function has to wait until the
*   V-sync has occured and the pending buffer is used by the display controller.
*   In case of an external display controller, the function has to wait until
*   the transfer (update) of the graphics data has been completed and there are
*   no pending buffers.
*
* ARGUMENTS:
*   None
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspDisplayWaitForCompletion( void )
{
  #if EW_USE_DOUBLE_BUFFER == 1

    CPU_LOAD_SET_IDLE();

    #if EW_USE_OPERATING_SYSTEM == 1

      /* wait until pending framebuffer is used as current framebuffer and
         use CPU time for other tasks */
      while ( PendingFramebuffer )
      {
        if ( EwBspOsSemaphoreWait( LcdUpdateSemaphore, 1000 ))
          EwBspOsSemaphoreRelease( LcdUpdateSemaphore );
      }

    #else

      /* wait until pending framebuffer is used as current framebuffer */
      while( PendingFramebuffer )
        ;

    #endif

    CPU_LOAD_SET_ACTIVE();

  #endif
}


/*******************************************************************************
* FUNCTION:
*   EwBspDisplayCommitBuffer
*
* DESCRIPTION:
*   The function EwBspDisplayCommitBuffer is called from the Graphics Engine
*   when the rendering of a certain buffer has been completed.
*   The type of buffer depends on the selected framebuffer concept.
*   If the display is running in a double-buffering mode, the function is called
*   after each buffer update in order to change the currently active framebuffer
*   address. Changing the framebuffer address should be synchronized with V-sync.
*   If the system is using an external graphics controller, this function is
*   responsible to start the transfer of the framebuffer content.
*
* ARGUMENTS:
*   aAddress - Address of the framebuffer to be shown on the display.
*   aX,
*   aY       - Origin of the area which has been updated by the Graphics Engine.
*   aWidth,
*   aHeight  - Size of the area which has been updated by the Graphics Engine.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspDisplayCommitBuffer( void* aAddress, int aX, int aY, int aWidth, int aHeight )
{
  #if EW_USE_DOUBLE_BUFFER == 1

    R_BSP_IrqDisable( DisplayConfig.line_detect_irq );

    /* set pending framebuffer address to be used on next V-sync */
    PendingFramebuffer = (uint32_t)aAddress;
    R_GLCDC_BufferChange( &g_display0_ctrl, (uint8_t *)aAddress, DISPLAY_FRAME_LAYER_1 );

    #if EW_USE_OPERATING_SYSTEM == 1
      EwBspOsSemaphoreWait( LcdUpdateSemaphore, 0 );
    #endif

    R_BSP_IrqEnable( DisplayConfig.line_detect_irq );

  #endif
}


/*******************************************************************************
* FUNCTION:
*   EwBspDisplaySetClut
*
* DESCRIPTION:
*   The function EwBspDisplaySetClut is called from the Graphics Engine
*   in order to update the hardware CLUT of the current framebuffer.
*   The function is only called when the color format of the framebuffer is
*   Index8 or LumA44.
*
* ARGUMENTS:
*   aClut - Pointer to a color lookup table with 256 entries.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspDisplaySetClut( unsigned long* aClut )
{
  display_clut_cfg_t clutCfg;
  clutCfg.p_base = aClut;
  clutCfg.start  = 0;
  clutCfg.size   = 256;

  R_GLCDC_Stop( &g_display0_ctrl );

  #if EW_USE_DOUBLE_BUFFER == 0
    /* in case of single-buffering a delay is necessary to ensure that CLUT is loaded */
    EwBspOsDelay(100);
  #endif

  /* load the colors into the hardware CLUT */
  R_GLCDC_ClutUpdate( &g_display0_ctrl, &clutCfg, DISPLAY_FRAME_LAYER_1 );
  R_GLCDC_Start( &g_display0_ctrl );
}

/* msy */
